#include "Quiz.h"

#include <fstream>
#include <iostream>
#include <string>
#include <set>
#include <stdlib.h>
#include <vector>
#include <algorithm>
#include <assert>

using namespace std;

class QuizLoader
{
    friend class Quiz;
    set <string, less<string> > questions;
    set <string, less<string> > answers;
    vector <pair <string, string > > words;
    string question1, question2, description, charset;
    int bins, askbothways;
public:
    bool ProcessFile (const string &fileName);
    void ProcessLine(string &line, int lineNum);
    QuizLoader();
};

QuizLoader::QuizLoader() : questions(), answers(), words(),
    question1("What is \"\"?"), question2("What is \"\"?"), description(""), charset("iso-latin-1"), bins(4), askbothways(1)
{
}

bool by_difficulty (const Word &a, const Word &b)
{
	return a.GetQuizCount() > b.GetQuizCount();
}

void Quiz::GetMostDifficult (vector <Word> &result, unsigned int amount)
{
	result.clear();
	sort (words.begin(), words.end(), by_difficulty);
	for (unsigned int i = 0; i < (amount < words.size() ? amount : words.size()); ++i)
	{
		result.push_back (words[i]);
	}
}


void QuizLoader::ProcessLine (string &line, int lineNum)
{
	// see if line contains only whitespace
	string first, last;
	if (line[0] == '#')
	{
		unsigned int split = line.find ("=");
		if (split != string::npos)
		{
			first = line.substr (1, split-1);
			last = line.substr (split + 1, line.size() - split - 1);
  
		    if (first.compare ("question1") == 0) { question1 = last; }
		    else if (first.compare ("question2") == 0) { question2 = last; }
		    else if (first.compare ("description") == 0) { description = last; }
		    else if (first.compare ("charset") == 0) { charset = last; }
		    else if (first.compare ("bins") == 0) { bins = atoi (last.c_str()); }
		    else if (first.compare ("askbothways") == 0) { askbothways = atoi (last.c_str()); }
		}
	}	
	else
 	{
		// see if line contains a comma + space
		unsigned int split = line.find (", ");
		if (split != string::npos)
		{
		    first = line.substr (0, split);
		    last = line.substr (split + 2, line.size() - split - 2);
		    // see if question or answer was already in the list earlier
		    if (questions.find (first) == questions.end())
		    {
			questions.insert (first);
			if (answers.find (last) == answers.end())
			{
			    answers.insert (last);
			    // save pair of words
			    words.push_back (pair<string, string>(first, last));
			}
			else
			{
			    cout << "Warning: duplicate answer \"" << last << "\" in line " << lineNum << " ignored." << endl;		
			}
		    }
		    else
		    {
			cout << "Warning: duplicate question \"" << first << "\" in line " << lineNum << " ignored." << endl;		
		    }
		}	
		else
		{
		    // check if line contains anything but whitespace
		    if (line.find_first_not_of ("\n\r\t ") != string::npos)
		    {
			cout << "Warning: line " << lineNum << ": \"" << line << "\" not OK" << endl;
		    }
		}
    }	
}

bool QuizLoader::ProcessFile (const string &fileName)
{
    
    int lineNum = 1;
    char c;
    string s;
    bool was015 = false;
    ifstream f;
    bool error = false;
    f.open (fileName.c_str(), ios::in | ios::binary);
    if (!f)
    {
	error = true;
    }
    while (!error)
    {
	// read until you get to /012 or /015. A 012 after a 015 are together a single line end.
	f.read (&c, sizeof (c));
	if (f.eof())
	{
	    // process remainder.
	    ProcessLine (s, lineNum++);	
	    break;
	}
	if (c == 012 || c == 015)
	{			
	    if (!(c == 012 && was015)) 
	    {
		    ProcessLine (s, lineNum++);
		    s = "";
	    }
	}
	else
	{
	    s += c;
	}
	was015 = (c == 015);									
    }
    f.close();
    
    return (error == false);
}

Quiz::Quiz()
{
    Reset();
}

void Quiz::Reset()
{
    words.clear();    
    wordmap.clear();
    fileName = "";
    fileTimeStamp = 0;
    totalDuration = 0;
    sessions = 0;
    counter = 1;
    for (int i = 0; i < MAXBINS; ++i) binCount[i] = 0;
    isValid = false; // gets valid after loading a quiz
    currentWord = words.end();
}

bool Quiz::IsFinished() const
{
    // check if all bins before last are empty
    bool result = true;
    for (int i = 0; i < GetBins()-1; ++i)
	{
		if (GetBinCount(i) != 0) result = false;
	}
    return result;
}

void Quiz::NextQuestion()
{
    random_shuffle (words.begin(), words.end());
    
    // go through list
    
    // 1. look for word that has highest due. If due > 3, pick it. If due > 2, pick it with 30% chance
    // 2. if none found yet, pick two words that are not in last bin. Return the one in lowest bin.
    
    vector <Word>::iterator i, j;
    
    currentWord = words.end();
    int maxDue = -1;
    for (i = words.begin(); i != words.end(); ++i)
    {
        if (i->GetHowSoon() != -1 && counter - i->GetHowSoon() > maxDue)
        {
            maxDue = counter - i->GetHowSoon();
            currentWord = i;
        }
    }
    
    // debugging
    //if (currentWord == words.end())
    //    cout << "--no due word found" << endl;
    //else
    //    cout << "--duest word ["<< maxDue <<"] is: " << currentWord->GetQuestion() << endl;
    
    int dueWait = 3;
    if (rand() % 100 < 40) dueWait = 2;
    //cout << "-- dueWait = " << dueWait << endl;
    
    // if no currentWord found, or there is no word with due higher than 3,
    // or 30% chance if there is no word higher than 2
    if  (currentWord == words.end() || maxDue < dueWait)       
    {
        i = words.begin();
        while (i != words.end() && (i->GetBin() >= bins-1 || i->GetHowSoon() != -1)) { i++; }        
        j = i;
        if (i != words.end())
        {
            while (j != words.end() && (j->GetBin() >= bins-1 || j->GetHowSoon() != -1)) { j++; }       
        }
        if (i != words.end() && j != words.end())
        {
            currentWord = i->GetBin() > j->GetBin() ? j : i;
        }       
    }
}

// when this is used, question counter is increased
// so call this only once per question/answer
bool Quiz::CompareAnswer (std::string &a)
{
    assert (currentWord != words.end());
    bool result = currentWord->CompareAnswer (a, counter, binCount);
    hint = "";
    if (!result)
    {
        map <string, string, less<string> >:: iterator f = wordmap.find (a);
        if (f != wordmap.end())
        {
            hint = "You may be confused with \"" + f->second + "\" -> \"" + f->first + "\"";
        }
    }
    counter++;
    return result;
}

string Quiz::GetCorrectAnswer ()
{
    assert (currentWord != words.end());
    return currentWord->GetAnswer();
}

string Quiz::GetQuestion ()
{
    assert (currentWord != words.end());
    string w = currentWord->GetQuestion();
    string q = currentWord->GetDir() == 0 ? question1 : question2;
    unsigned int pos = q.find ("\"\"");
    if (pos != string::npos)
    {
		return q.substr (0, pos+1) + w + q.substr (pos + 1, q.size() - pos - 1);
    }
    else
    {
		return q + w;
    }
}

bool Quiz::HasHint () const
{
    return hint.size() > 0;
}

// load a lesson file
// return false when an error occurs.
bool Quiz::LoadLesson (const string &newFileName)
{
    bool isOk = true;
    
    QuizLoader ql;
    
    isOk = ql.ProcessFile (newFileName);    
    
    Reset();
    
    fileName = newFileName;
    question1 = ql.question1;
    question2 = ql.question2;
    description = ql.description;
    charset = ql.charset;
    bins = ql.bins;
    askbothways = ql.askbothways;

    // create words
    vector<pair<string, string> >::iterator i;
    int l = 0;
    for (i = ql.words.begin(); i != ql.words.end(); ++i)
    {
        words.push_back (Word (i->first, i->second, 0, l));
        wordmap[i->first] = i->second;
        if (askbothways) 
        {
            words.push_back (Word (i->second, i->first, 1, l));
            wordmap[i->second] = i->first;
        }
    }
    binCount[0] = words.size();
    random_shuffle (words.begin(), words.end());
    currentWord = words.begin();
    
    if (isOk) isValid = true;    
    return isOk;
}

bool Quiz::Deserialize (fstream &f)
{
    Reset();
    question1 = ReadString (f);
    question2 = ReadString (f);
    description = ReadString (f);
    charset = ReadString (f);
    bins = ReadInt (f);
    askbothways = ReadInt (f);
    int size = ReadInt (f);
    for (int i = 0; i < size; ++i)
    {
	string temp;
	Word w = Word(temp, temp, 0, 0);
	w.Deserialize (f);
	words.push_back (w);
	wordmap [w.GetQuestion()] = w.GetAnswer();
    }
    currentWord = words.end();
    
    fileName = ReadString (f);
    fileTimeStamp = ReadInt (f);
    totalDuration = ReadInt (f);
    sessions = ReadInt (f);
    counter = ReadInt (f);
    
    for (int k = 0; k < MAXBINS; ++k)
    {
	binCount[k] = ReadInt (f);
    }
    isValid = true;
    return true;
}

bool Quiz::Serialize (fstream &f) const
{
    WriteString (f, question1);
    WriteString (f, question2);
    WriteString (f, description);
    WriteString (f, charset);
    WriteInt (f, bins);
    WriteInt (f, askbothways);
    WriteInt (f, words.size());
    vector <Word>::const_iterator i;
    for (i = words.begin(); i != words.end(); ++i)
    {
        i->Serialize (f);
    }
    WriteString (f, fileName);
    WriteInt (f, fileTimeStamp);
    WriteInt (f, totalDuration);
    WriteInt (f, sessions);
    WriteInt (f, counter);
    
    // note: string hint not written
    // note: Word *currentWord not written
    for (int k = 0; k < MAXBINS; ++k)
    {
        WriteInt (f, binCount[k]);
    }
    
    return true;
}

void Quiz::SetBins (int newCount)
{
    bins = newCount;
}

    